import os
import copy

import torch
import matplotlib.pyplot as plt
from knnslim.graph import model_converter
from knnslim.utils import enums, pruner_utils


class ADMMBaseCallback:
    def __init__(self, admm_pruner, ckpt_dir='checkpoints', stage='pretrain'):
        self.admm_pruner = admm_pruner
        self.ckpt_dir = ckpt_dir
        self.stage = stage

    def on_train_begin(self):
        if self.admm_pruner.resume:
            pruner_utils.load_breakpoint(self.admm_pruner)
        elif self.admm_pruner.load_model:
            pruner_utils.load_pretrained_model(self.admm_pruner)

    def on_epoch_begin(self, epoch):
        self.admm_pruner.logger.info(f'Epoch: {epoch}')

    def on_batch_begin(self):
        pass

    def on_batch_end(self):
        pass

    def on_epoch_end(self, epoch, optimizer):
        ckpt_path = f'./{self.ckpt_dir}/pid_{self.admm_pruner.pid}_{self.stage}_epoch_{epoch}.pt'
        pruner_utils.save_checkpoint(self.admm_pruner, epoch, optimizer, ckpt_path)

    def on_train_end(self):
        ckpt_path = f'./{self.ckpt_dir}/pid_{self.admm_pruner.pid}_{self.stage}.pt'
        pruner_utils.publish(self.admm_pruner, ckpt_path)

    def on_admm_loss(self):
        return torch.zeros([], device=self.admm_pruner.device)


class ADMMPretrainCallback(ADMMBaseCallback):
    def __init__(self, admm_pruner):
        super().__init__(admm_pruner, stage='pretrain')


class ADMMPruneCallback(ADMMBaseCallback):
    def __init__(self, admm_pruner):
        super().__init__(admm_pruner, stage='prune')
        self.epochs = []
        self.admm_losses = []
        self.result_dir = './admm_loss_figure'
        if not os.path.exists(self.result_dir):
            os.makedirs(self.result_dir, exist_ok=True)

    def on_train_begin(self):
        super().on_train_begin()
        self.admm_pruner.logger.info('Initializing ADMM variables: Z and U')
        self.admm_pruner.initialize_z_u()

    def on_epoch_begin(self, epoch, optimizer=None):
        super().on_epoch_begin(epoch)
        self.admm_pruner.update_z_u(epoch)
        lr = self.admm_pruner.adjust_learning_rate(optimizer, epoch)
        self.admm_pruner.logger.info(f'Current learning rate: {lr}')

    def on_epoch_end(self, epoch, optimizer):
        super().on_epoch_end(epoch, optimizer)
        admm_loss = self.admm_pruner.get_admm_loss().item()
        self.admm_pruner.logger.info(f'ADMM loss: {admm_loss}')
        self.epochs.append(epoch)
        self.admm_losses.append(admm_loss)

    def on_train_end(self):
        fig, ax = plt.subplots()
        ax.plot(self.epochs, self.admm_losses, label=f'rho_{self.admm_pruner.rho}_admm_epoch_{self.admm_pruner.admm_epoch}')
        ax.set_ylabel('ADMM loss')
        ax.set_xlabel('Epoch')
        ax.legend()
        plt.savefig(f'./{self.result_dir}/pid_{self.admm_pruner.pid}_{self.stage}_admm_loss.jpg')

        super().on_train_end()

    def on_admm_loss(self):
        return self.admm_pruner.get_admm_loss()


class ADMMRetrainCallback(ADMMBaseCallback):
    def __init__(self, admm_pruner):
        super().__init__(admm_pruner, stage='retrain')

    def on_train_begin(self):
        """
            1. generate dependencies of all weights in the model
            2. hard prune each target weight, return their masks
            3. synchronize masks related to 'Add' of 'Concat'
            4. define prune ops, shrink ops
        """
        ori_model = copy.deepcopy(self.admm_pruner.model)
        super().on_train_begin()
        if not self.admm_pruner.resume:  # 非断点续训，大模型需要转成小模型
            self.admm_pruner.get_weight_masks()
            self.admm_pruner.sync_add_cat_masks()
            self.admm_pruner.merge_all_masks()
            self.admm_pruner.logger.info('Start to prune all weights')
            self.admm_pruner.prune_all_weights()

            if self.admm_pruner.pruning_method == enums.PruningMethod.Structured:
                sm_converter = model_converter.SmallModelConverter(self.admm_pruner.model, self.admm_pruner.analyzer)
                small_model = sm_converter.converter(self.admm_pruner.masks)
                torch.save(small_model, f'slim_model_{self.admm_pruner.model.name}.pt')
                self.admm_pruner.model = small_model.to(self.admm_pruner.device)

        if self.admm_pruner.num_devices > 1:
            self.admm_pruner.model = torch.nn.DataParallel(self.admm_pruner)
        pruner_utils.get_prune_summary(ori_model, self.admm_pruner.model, 
                                       self.admm_pruner.inputs_tuple,
                                       self.admm_pruner.logger)
        del ori_model
