"""
========================================================================
Copyright 2022 Pan Siyuan, Hou Liang, Qiu Kaibin, Xiang Hao

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
=========================================================================

ADMM Pruning Algorithm.
"""
from __future__ import print_function
import json
import sys

import torch

from knnslim.callback import pruner_callback
from knnslim.prune import pruning_method
from knnslim.prune import pruning_criterion
from knnslim.utils import enums
from knnslim.utils import pruner_utils
from knnslim.graph import model_analyzer


class ADMMPruner:
    """
    ADMM model pruning algorithm
    """

    def __init__(self, model, config, stage, inputs_tuple, pid=0, load_model=False, resume=False, pretrained_state=None,
                 external_init_lr=None):
        """Initializes ADMMPruner with config

        Args:
            model(pytorch module): object of torch model.
            config(json file): path to the *.json/yaml file, which contains detailed prune ratios for the model weights
            stage(str): ADMM optimization stage, 'prune' or 'fine-tune'.
            pid(int): indicates task id.
            input_shape(tuple): input data shape，to trace the torch model.
            external_init_lr(float): external_init_lr(float): If the init_lr is passed in externally, the init_lr
            becomes external_lr instead.
            resume(bool): if resume is True, model needs load ckpt to continue training.
        """
        # Auxiliary parameters in the ADMM algorithm process
        self.model = model
        self.inputs_tuple = inputs_tuple
        self.stage = enums.str2enums_id(stage)
        self.pid = pid
        self.admm_u = {}
        self.admm_z = {}
        self.masks = {}
        self.logger = pruner_utils.get_logger(pid, stage)
        self.num_devices = pruner_utils.get_num_device(model)
        self.device = torch.device('cpu' if self.num_devices == 0 else 'cuda')
        self.logger.info(f'------------- ADMM {self.stage} -------------')
        # ====================================
        self.load_model = load_model
        self.resume = resume
        self.pretrained_state = pretrained_state
        self.start_epoch, self.optimizer = 0, None
        # ====================================

        # Model analysis
        self.train_weights = dict(self.model.named_parameters())
        self.analyzer = model_analyzer.ModelAnalyzer(model, inputs=inputs_tuple)
        self.weight_to_module = self.analyzer.weight_to_module

        # The hyperparameter that the ADMM algorithm needs to configure, reading from json configuration file
        json_cfg = self._load_config(config)
        self.pruning_method = enums.str2enums_id(json_cfg['pruning_method'])
        self.pruning_criterion = enums.str2enums_id(json_cfg['pruning_criterion'])
        self.pruning_engine = enums.str2enums_id(json_cfg['pruning_engine'])
        self.rho = json_cfg['rho']
        self.admm_epoch = json_cfg['admm_epoch']
        self.internal_stage = json_cfg['internal_stage']
        self.init_lr = json_cfg['init_lr'] if external_init_lr is None else external_init_lr
        self.lr_decay = json_cfg['lr_decay']
        self.base_layer_pruning_ratio = json_cfg['base_layer_pruning_ratio']
        self.special_layer_pruning_ratio = json_cfg['special_layer_pruning_ratio']
        self.exclude_layer_pruning_ratio = json_cfg['exclude_layer_pruning_ratio']
        self.extra_dependencies = json_cfg['extra_dependencies']

        # 根据配置解析获得每层的剪枝率
        self.prune_ratios = self._parse_prune_ratios()
        pruner_utils.generate_yaml(self.train_weights, json_cfg, self.prune_ratios, config)

        self.prune_weights = self._get_prune_weights()
        self.pruner, self.criterion = self._get_prune_method()
        self.callback = self._get_admm_callback()
        self.dependencies = self._create_dependencies()  # 获得add, cat等算子的前后依赖

    def _load_config(self, config):
        """Loads the configuration(json) file and returns it as a dictionary"""
        self.logger.info('Loading admm config(json) ...')
        with open(config, 'r', encoding='utf-8') as f:
            return json.load(f)

    def _get_prune_weights(self):
        """Gains layers(to be pruned) and corresponding weights according to prune_ratios.

        Return:
            prune_weights(dict): layer(to be pruned) name and corresponding weights.
        """
        prune_weights = {}
        for name, param in self.train_weights.items():
            if name in self.prune_ratios:
                prune_weights[name] = param
        return prune_weights

    def _get_prune_method(self):
        """
        Select one prune method according to the pruning_method, pruning_criterion, pruning_engine
        """
        criterion = None
        if self.pruning_criterion == enums.PruningCriterion.L1:
            criterion = pruning_criterion.LnNormPruningCriterion(p=1, pruning_engine=self.pruning_engine)
        elif self.pruning_criterion == enums.PruningCriterion.L2:
            criterion = pruning_criterion.LnNormPruningCriterion(p=2, pruning_engine=self.pruning_engine)
        elif self.pruning_criterion == enums.PruningMethod.Unstructured:
            criterion = pruning_criterion.UnstructuredPruningCriterion(self.pruning_engine)

        pruner = None
        if self.pruning_method == enums.PruningMethod.Filter:
            pruner = pruning_method.FilterPruningMethod(criterion, self.pruning_engine)
        elif self.pruning_method == enums.PruningMethod.Channel:
            pruner = pruning_method.ChannelPruningMethod(criterion, self.pruning_engine)
        elif self.pruning_method == enums.PruningMethod.Unstructured:
            pruner = pruning_method.UnstructuredPruningMethod(criterion, self.pruning_engine)
        elif self.pruning_method == enums.PruningMethod.Bcr:
            pruner = pruning_method.BCRPruningMethod(criterion, self.pruning_engine)

        return pruner, criterion

    def _parse_prune_ratios(self):
        """Obtains the pruning rate configuration file of each layer of the model,
        according to the model structure and pruning rate configuratrion.
        """
        self.logger.info('Obtaining the pruning rate of each layer')
        prune_ratios = {}
        for name, _ in self.train_weights.items():
            if 'weight' in name:
                layer_type = pruner_utils.get_type(self.weight_to_module[name])
                if layer_type == 'Conv2D':
                    if 'Conv2D' in self.base_layer_pruning_ratio:
                        prune_ratios[name] = self.base_layer_pruning_ratio['Conv2D']
                elif layer_type == 'ConvTranspose2d':
                    if 'ConvTranspose2d' in self.base_layer_pruning_ratio:
                        prune_ratios[name] = self.base_layer_pruning_ratio['Conv2D']
                # elif layer_type == 'DepthwiseConv2D':
                #     if 'DepthwiseConv2D' in self.base_layer_pruning_ratio:
                #         prune_ratios[name] = self.base_layer_pruning_ratio['DepthwiseConv2D']
                elif 'Linear' in layer_type:
                    if 'Linear' in self.base_layer_pruning_ratio:
                        prune_ratios[name] = self.base_layer_pruning_ratio['Linear']

        self.special_layer_pruning_ratio = {pruner_utils.prefix(k, self.num_devices): v for k, v in
                                            self.special_layer_pruning_ratio.items()}
        self.exclude_layer_pruning_ratio = {pruner_utils.prefix(k, self.num_devices) for k in
                                            self.exclude_layer_pruning_ratio}

        for name, value in self.special_layer_pruning_ratio.items():
            prune_ratios[name] = value

        for name in self.exclude_layer_pruning_ratio:
            if name in prune_ratios:
                prune_ratios.pop(name)

        return prune_ratios

    def _create_dependencies(self, input_shape=None):
        self.logger.info(f'-> Automatically generate [{self.model.name}] dependencies')
        # trace graph to obtain model dependencies
        dependencies = self.analyzer.create_auto_dependencies(input_shape)

        # load extra dependencies from the yaml file
        yaml_file = getattr(self, 'dependency_file', None)
        extra_dependencies = pruner_utils.load_extra_dependencies(yaml_file, self.num_devices)
        dependencies = pruner_utils.merge_dependencies(dependencies, extra_dependencies)
        dependencies = pruner_utils.revise_dependencies(dependencies, self.exclude_layer_pruning_ratio)
        dependencies = pruner_utils.gain_cat_prev_dependencies(dependencies)
        dependencies = pruner_utils.detach_cat_add_dependencies(dependencies)  # 将cat-add算子的依赖从cat剥离出来

        dependency_info = ['---------------------', 'dependencies[weight] :']
        for k, v in dependencies['weight'].items():
            dependency_info.append(f'{k}:\n\t{v}')

        dependency_info.extend(['---------------------', 'dependencies[add] :'])
        for k, v in dependencies['add'].items():
            dependency_info.append(f'{k}:\n\t{v}')

        dependency_info.extend(['---------------------', 'dependencies[cat] :'])
        for k, v in dependencies['cat'].items():
            dependency_info.append(f'{k}:\n\t{v}')
        # self.logger.info('\n'.join(dependency_info)) # 打印依赖

        return dependencies

    def calculate_cat_virtual_weight_and_ratio(self):
        """计算cat算子的虚拟权重和虚拟剪枝率。
        虚拟权重：cat算子前依赖的权重cat的结果
        虚拟剪枝率：cat算子前依赖的权重剪枝率对输出通道数的加权
        """
        all_cat_info = {}
        for cat_name, weights in self.dependencies['cat_prev_dict'].items():
            cat_info = {}
            cat_weight = torch.tensor([], dtype=torch.float32, device=self.device)
            cat_channels = 0
            total_channels = 0
            for weight in weights:
                params = self.prune_weights[weight].detach()
                cat_weight = torch.cat([cat_weight, params], dim=0)
                cat_channels += self.prune_ratios[weight] * params.shape[0]
                total_channels += params.shape[0]
            # 根据cat前依赖给cat算子一个虚拟的剪枝率
            cat_module = 'cat_op'
            cat_prune_ratio = cat_channels / total_channels
            cat_info['virtual_weight'] = cat_weight
            cat_info['virtual_ratio'] = cat_prune_ratio
            cat_info['virtual_module'] = cat_module
            all_cat_info[cat_name] = cat_info
        return all_cat_info

    def initialize_z_u(self):
        self.logger.info('--- ADMM Z/U Initialize ---')
        for name, param in self.prune_weights.items():
            self.admm_u[name] = torch.zeros(param.shape, device=self.device)
            self.admm_z[name] = param.detach().clone().to(self.device)
        self._initialize_or_update_z(flag=False)

    def update_z_u(self, epoch, batch_idx=0):
        """Update Z and U, according to ADMM optimization with the pruning method

        Args:
            epoch(int): number of epochs in the main loop, starts from 0
            admm_epoch(int): interval epoch to update Z and U
            batch_idx(int): index of the batch size
        """
        if epoch > 0 and epoch % self.admm_epoch == 0 and batch_idx == 0:
            self.logger.info('--- ADMM Z Update ---')
            self._initialize_or_update_z(flag=True)
            self.logger.info('--- ADMM U Update ---')
            self._update_u()

    def _initialize_or_update_z(self, flag=False):
        """Initializes or updates Z variables

        Args:
            flag(bool): This flag indicates whether update or initialize. Optional items:
                True: for updating Z
                False: for initializing Z
        """
        for name, param in self.prune_weights.items():
            if (name not in self.dependencies['add_prev_set']) and (
                    name not in self.dependencies['cat-add_prev_set']):  # normal prune to weights
                # initialize: Z(0) = W(0), update: Z(k+1) = W(k+1) + U(k)
                z_next = param.detach() + self.admm_u[name] if flag else param.detach()
                z_next_module = self.weight_to_module[name]  # 获得该层权重名

                # update Z using the pruning method
                updated_z, _ = self.pruner.prune(z_next_module, z_next, self.prune_ratios[name])
                pruner_utils.assign_tensor(self.admm_z[name], updated_z)

        cat_virtual_info = self.calculate_cat_virtual_weight_and_ratio()
        for prev_add in self.dependencies['add']:  # sync prune to add conv weights
            z_next_list = []  # Z(k+1) list
            z_next_module_list = []
            ratios = []
            for name in prev_add:
                weight = None
                ratio = 0
                module = ''
                if 'cat' in name:  # add前面是不是普通权重是cat算子
                    weight, ratio, module = cat_virtual_info[name]['virtual_weight'], cat_virtual_info[name][
                        'virtual_ratio'], cat_virtual_info[name]['virtual_module']
                    self.prune_weights[name] = weight  # 给予cat算子一个虚拟权重，在计算admm_loss时需要跳过
                    self.admm_z[name] = weight  # 给予cat算子一个虚拟的权重
                    self.admm_u[name] = weight
                elif name in self.prune_weights:
                    weight = self.prune_weights[name].detach().clone()  # 对于每个cat-add的权重在每次剪枝都需要重新计算，这是一个问题
                    ratio = self.prune_ratios[name]
                    module = self.weight_to_module[name]

                # initialize: Z(0) = W(0), update: Z(k+1) = W(k+1) + U(k)
                z_next = weight.detach().clone() + self.admm_u[name] if flag else weight.detach().clone()
                z_next_module = module

                z_next_list.append(z_next)
                z_next_module_list.append(z_next_module)
                ratios.append(ratio)
            assert len(set(ratios)) == 1, f'prune_ratios {ratios} in {prev_add} are different'

            # update Z using the prune method
            # self.logger.info(f'\nprev_add:\n{prev_add}')
            updated_values, _ = self.pruner.sync_prune(z_next_module_list, z_next_list, ratios)
            for name, updated_z in zip(prev_add, updated_values):
                pruner_utils.assign_tensor(self.admm_z[name], updated_z)

        for cat_op, prev_weights in self.dependencies['cat_prev_dict'].items():
            # 对cat-add之前的权重求mask后的admm_z
            if cat_op in self.dependencies['cat-add_set']:
                start = 0
                for weight in prev_weights:
                    channel_num = self.prune_weights[weight].shape[0]
                    pruner_utils.assign_tensor(self.admm_z[weight], self.admm_z[cat_op][start: start + channel_num])
                    start = start + channel_num

    def _update_u(self):
        """update U variables"""
        for name, param in self.prune_weights.items():
            weight = param.detach() if self.device == torch.device('cuda') else param.detach().cpu()
            # U(k+1) = W(k+1) - Z(k+1) + U(k)
            u_tensor = weight - self.admm_z[name] + self.admm_u[name]
            pruner_utils.assign_tensor(self.admm_u[name], u_tensor)

    def get_admm_loss(self):
        """
            Compute the L2 loss of ADMM, append to the original loss
        :return:
            admm_loss (Tensor): to be added to the main loss
        """
        admm_loss = torch.tensor(0., device=self.device)
        for name, param in self.prune_weights.items():
            if 'cat' in name:
                continue
            weight = param if self.device == torch.device('cuda') else param.cpu()  # require gradient
            l2_norm = torch.norm(weight - self.admm_z[name] + self.admm_u[name], p=2) ** 2
            admm_loss += 0.5 * self.rho * l2_norm

        return admm_loss

    def adjust_learning_rate(self, optimizer, epoch):
        """Re-adjusts the learning rate in the ADMM optimization. Learning rate change is periodic, decay 3 times.
        When epoch is dividable by admm_epoch, learning rate is reset to the original value.

        Args:
            optimizer(torch.optim): optimizer to adjust lr during training
            epoch(int): number of epochs in the main loop
            default value from config.
            self.init_lr(float): initial learning rate
            self.internal_stage(int): decay times of learning rate in one admm_epoch
            self.lr_decay(float): multiple to the learning rate in each admm_step
        Return:
            real learning rate.
        """
        if epoch % self.admm_epoch == 0:
            lr = self.init_lr
        else:
            offset = epoch % self.admm_epoch
            admm_step = self.admm_epoch / self.internal_stage
            lr = self.init_lr * (self.lr_decay ** (offset // admm_step))

        if isinstance(optimizer, dict):
            for key in optimizer:
                for param_group in optimizer[key].param_groups:
                    param_group['lr'] = lr
        else:
            for param_group in optimizer.param_groups:
                param_group['lr'] = lr

        return lr

    def prune_all_weights(self):
        """Hard prune on each weight of the model

        Args:
            weights: list of weights to be pruned
            masks: list of masks to use
        """
        for name, param in self.train_weights.items():
            if name in self.masks:
                pruner_utils.assign_tensor(param, param * self.masks[name])

    def merge_all_masks(self):

        weights = self.prune_weights
        self.logger.info('--- 1. 获得依赖非add/cat层的层mask ---')
        for prev_var, _ in weights.items():
            if prev_var not in self.dependencies['weight']:
                self.logger.info(f'{prev_var} is {self.weight_to_module[prev_var]}, not belongs to active prune')
                continue
            for pass_var_name in self.dependencies['weight'][prev_var]:
                # for prev_var, pass_vars in self.dependencies['weight'].items():
                #     for pass_var_name in pass_vars:
                passive_mask = self._get_passive_masks(pass_var_name, self.masks[prev_var],
                                                       self.weight_to_module[prev_var])
                # 融合active和passive的mask
                var_name = pass_var_name
                if 'running_mean' in pass_var_name or 'running_var' in pass_var_name:
                    var_name = f'{".".join(pass_var_name.split(".")[:-1])}.weight'

                active_mask = self.masks[var_name] if var_name in self.masks else torch.ones_like(
                    self.train_weights[var_name])
                self.masks[pass_var_name] = active_mask * passive_mask

        self.logger.info('--- 2. 获得依赖add层的层mask ---')
        for _, (add_list, next_add) in self.dependencies['add'].items():
            # 获得add的mask
            add_mask = self.masks[add_list[0]]
            for pass_var_name in next_add:
                passive_mask = self._get_passive_masks(pass_var_name, add_mask)
                var_name = pass_var_name
                if 'running_mean' in pass_var_name or 'running_var' in pass_var_name:
                    var_name = f'{".".join(pass_var_name.split(".")[:-1])}.weight'

                active_mask = torch.ones_like(self.train_weights[var_name])
                if var_name in self.masks:
                    active_mask = self.masks[var_name]
                self.masks[pass_var_name] = active_mask * passive_mask

        self.logger.info('--- 3. 获得依赖cat层的层mask ---')
        for _, (cat_list, next_cat) in self.dependencies['cat'].items():
            cat_mask = self.masks[cat_list[0]]
            for pass_var_name in next_cat:
                passive_mask = self._get_passive_masks(pass_var_name, cat_mask)
                var_name = pass_var_name
                if 'running_mean' in pass_var_name or 'running_var' in pass_var_name:
                    var_name = f'{".".join(pass_var_name.split(".")[:-1])}.weight'

                active_mask = torch.ones_like(self.train_weights[var_name])
                if var_name in self.masks:
                    active_mask = self.masks[var_name]
                self.masks[pass_var_name] = active_mask * passive_mask

    def _get_passive_masks(self, var_name, dependency_mask, dependency_module=None):
        """根据上一层的输出通道变化(上一层的mask)对本层输入通道进行被动剪枝，得到输入通道的mask。
        Handles passive pruning of modules whose number of input channels is restricted by the output channels
        of the current module
        Args:
            var_name:  weight name to be passive pruned
            dependency_mask: 制约层/被依赖层的权重mask
        """
        if isinstance(self.pruner, pruning_method.UnstructuredPruningMethod):
            return None
        dependency_mask = torch.transpose(dependency_mask, 0, 1) \
            if isinstance(dependency_module, torch.nn.ConvTranspose2d) else dependency_mask
        mask = dependency_mask.squeeze()
        mask_2d = mask.reshape(mask.shape[0], -1)
        filter_, column = mask_2d.shape[0], mask_2d.shape[1]

        mask_1d = None
        if self.pruning_method == enums.PruningMethod.Channel:
            mask_1d = torch.mean(mask_2d, dim=0)
        elif self.pruning_method == enums.PruningMethod.Filter:
            mask_1d = torch.mean(mask_2d, dim=1)

        # BatchNorm2d
        if 'running_mean' in var_name or 'running_var' in var_name:
            return mask_1d

        cur_var_shape = self.train_weights[var_name].shape  # 当前层参数shape
        cur_var_layer = self.weight_to_module[var_name]  # 当前层对应的模块
        cur_var_type = pruner_utils.get_type(cur_var_layer)  # 当前层所属模块类型
        auxiliary_mask = torch.ones(cur_var_shape, device=self.device, dtype=torch.float32)

        mask_nd = None
        # 1D, Batchnorm, bias, etc
        if len(cur_var_shape) == 1:
            mask_nd = mask_1d
        # FC
        elif len(cur_var_shape) == 2:
            if self.pruning_method == enums.PruningMethod.Channel:
                mask_nd = mask_1d.reshape((column, 1)) * auxiliary_mask
            elif self.pruning_method == enums.PruningMethod.Filter:
                mask_nd = mask_1d.reshape((1, filter_)) * auxiliary_mask

        # Conv Layer
        elif len(cur_var_shape) == 4:
            if self.pruning_method == enums.PruningMethod.Channel:
                if cur_var_type == 'DepthwiseConv2D':
                    mask_nd = mask_1d.reshape((1, column, 1, 1)) * auxiliary_mask
                elif cur_var_type == 'Conv2D':
                    mask_nd = mask_1d.reshape((column, 1, 1, 1)) * auxiliary_mask
            elif self.pruning_method == enums.PruningMethod.Filter:
                if cur_var_type == 'Conv2D':
                    mask_nd = mask_1d.reshape((1, filter_, 1, 1)) * auxiliary_mask
                elif cur_var_type == 'DepthwiseConv2D':
                    mask_nd = mask_1d.reshape((filter_, 1, 1, 1)) * auxiliary_mask
                elif cur_var_type == 'ConvTranspose2d':
                    mask_nd = mask_1d.reshape(filter_, 1, 1, 1) * auxiliary_mask
        else:
            self.logger.info(f'Var.shape.ndims has to be 1,2 or 4, but got {len(cur_var_shape)}')
            sys.exit(1)

        return mask_nd

    def sync_add_cat_masks(self):
        """对于add/cat前依赖的待剪枝层的输出通道进行剪枝对应的mask"""
        masks = self.masks

        if isinstance(self.pruner, pruning_method.UnstructuredPruningMethod):
            pass
        # Add: sync prune conv weights in dependencies['add']
        cat_virtual_info = self.calculate_cat_virtual_weight_and_ratio()
        for prev_add, (add_list, _) in self.dependencies['add'].items():
            weight_list = []
            module_list = []
            ratios = []
            for name in prev_add:
                weight = None
                ratio = 0
                module = ''
                if 'cat' in name:  # add前面是不是普通权重是cat算子
                    weight, ratio, module = cat_virtual_info[name]['virtual_weight'], cat_virtual_info[name][
                        'virtual_ratio'], cat_virtual_info[name]['virtual_module']
                    self.prune_weights[name] = weight  # 给予cat算子一个虚拟权重，在计算admm_loss时需要跳过
                elif name in self.prune_weights:
                    weight = self.prune_weights[name].detach().clone()  # 对于每个cat-add的权重在每次剪枝都需要重新计算，这是一个问题
                    ratio = self.prune_ratios[name]
                    module = self.weight_to_module[name]

                weight_list.append(weight)
                module_list.append(module)
                ratios.append(ratio)
            assert len(set(ratios)) == 1, f'prune_ratios {ratios} in {prev_add} are different'
            _, add_masks = self.pruner.sync_prune(module_list, weight_list, ratios)
            for name, add_mask in zip(prev_add, add_masks):
                masks[name] = torch.Tensor.float(add_mask)

            # create 1D masks for add ops
            mask1d = None
            prev_name, add_mask = prev_add[0], add_masks[0]
            mask1d = add_mask.reshape(add_mask.shape[0], -1).to(torch.float32).mean(dim=1)
            mask1d[mask1d > 0] = 1.0

            for add_name in add_list:
                masks[add_name] = mask1d

        # Cat: combine previous masks in dependencies['cat']
        for prev_cat, (cat_list, _) in self.dependencies['cat'].items():
            cat_mask = torch.tensor([], dtype=torch.float32, device=self.device)
            # 注意：cat前依赖层没有相互依赖
            for var_name in prev_cat:
                var_mask = masks[var_name].squeeze()
                var_mask = torch.Tensor.float(var_mask.reshape(var_mask.shape[0], -1)).mean(axis=1)
                var_mask[var_mask > 0] = 1.0
                cat_mask = torch.cat((cat_mask, var_mask), dim=0)
            # create 1D masks for cat ops
            for cat_name in cat_list:
                masks[cat_name] = cat_mask

        for cat_op, prev_weights in self.dependencies['cat_prev_dict'].items():
            if cat_op not in self.dependencies['add_prev_set']:
                continue
            start = 0
            for weight in prev_weights:
                channel_num = self.prune_weights[weight].shape[0]
                weight_mask = masks[cat_op][start:start + channel_num]
                start = start + channel_num
                masks[weight] = weight_mask

    def get_weight_masks(self):
        """获得对于非add前依赖的待剪枝层的输出通道进行剪枝对应的mask,cat前依赖的所有层不互相依赖"""
        # self.logger.info('--- get all masks of target weights ---')
        weights = self.prune_weights
        for name, param in weights.items():
            if (name not in self.dependencies['add_prev_set']) and (
                    name not in self.dependencies['cat-add_prev_set']):  # normal prune to weights
                weight = param.detach().clone()
                module = self.weight_to_module[name]

                _, mask = self.pruner.prune(module, weight, self.prune_ratios[name])
                # save each mask for every pruned weight
                self.masks[name] = torch.Tensor.float(mask)

    def _get_admm_callback(self):
        callback = None
        if self.stage == enums.PruningStage.Pretrain:
            callback = pruner_callback.ADMMPretrainCallback(self)
        if self.stage == enums.PruningStage.Prune:
            callback = pruner_callback.ADMMPruneCallback(self)
        elif self.stage == enums.PruningStage.Retrain:
            callback = pruner_callback.ADMMRetrainCallback(self)
        return callback
