import numpy as np
import torch

from knnslim.utils import enums

# Network trimming: A data-driven neuron pruning approach towards efficient deep architectures
# An entropy-based pruning method for cnn compression
# Learning efficient convolutional networks through network slimming

class BasePruningCriterion:
    def __init__(self, pruning_type=None, pruning_engine=None):
        self._validate_pruning_type(pruning_type)
        self._validate_pruning_engine(pruning_engine)

        self.pruning_type = pruning_type
        self.pruning_engine = pruning_engine

    def _validate_pruning_type(self, pruning_type):
        assert pruning_type in (enums.PruningType.Structured, enums.PruningType.Unstructured)

    def _validate_pruning_engine(self, pruning_engine):
        assert pruning_engine in (enums.PruningEngine.PyTorch, enums.PruningEngine.NumPy)

    def __call__(self, module=None, weight=None):
        raise NotImplementedError


class StructuredPruningCriterion(BasePruningCriterion):
    def __init__(self, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_type=enums.PruningType.Structured, pruning_engine=pruning_engine)

class LnNormPruningCriterion(StructuredPruningCriterion):
    def __init__(self, p=1, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_engine=pruning_engine)
        self._validate_norm(p)
        self.p = p
    
    def _validate_norm(self, p):
        assert isinstance(p, int) and p > 0

    def __call__(self, module=None, weight=None):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            if (module and isinstance(module, torch.nn.Linear)) or len(weight.shape) == 2:
                importance_scores = torch.norm(weight, p=self.p, dim=(1,))
            elif (module and isinstance(module, torch.nn.Conv1d)) or len(weight.shape) == 3:
                importance_scores = torch.norm(weight, p=self.p, dim=(1,2))
            elif (module and isinstance(module, torch.nn.Conv2d)) or len(weight.shape) == 4:
                importance_scores = torch.norm(weight, p=self.p, dim=(1,2,3))
            elif (module and isinstance(module, torch.nn.Conv3d)) or len(weight.shape) == 5:
                importance_scores = torch.norm(weight, p=self.p, dim=(1,2,3,4))
            elif (module and isinstance(module, torch.nn.ConvTranspose1d)) or len(weight.shape) == 3:
                importance_scores = torch.norm(weight, p=self.p, dim=(1,2))
            elif (module and isinstance(module, torch.nn.ConvTranspose2d)) or len(weight.shape) == 4:
                importance_scores = torch.norm(weight, p=self.p, dim=(1,2,3))
            elif (module and isinstance(module, torch.nn.ConvTranspose3d)) or len(weight.shape) == 5:
                importance_scores = torch.norm(weight, p=self.p, dim=(1,2,3,4))
            return importance_scores
            # weight_matrix = weight.reshape(weight.shape[0], -1)
            # return torch.linalg.norm(weight_matrix, ord=self.p, dim=1)
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            weight_matrix = weight.reshape(weight.shape[0], -1)
            return np.linalg.norm(weight_matrix, ord=self.p, axis=1)


class UnstructuredPruningCriterion(BasePruningCriterion):
    def __init__(self, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_type=enums.PruningType.Unstructured, pruning_engine=pruning_engine)

    def __call__(self, module=None, weight=None):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return torch.abs(weight)
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return np.abs(weight)
