import math
import sys
sys.path.append('../../..')

import numpy as np
import torch

from knnslim import prune
from knnslim.utils import enums


class BasePruningMethod:
    __doc__ = r"""Base pruning method for weight sparsity.
    Args:
        pruning_criterion (enum.Enum): the criterion to calculate the importance of weights/filters/channels
        pruning_type (enum.Enum): Structured or Unstructure sparsity
        pruning_engine (enum.Enum): PyTorch or NumPy, the engine for weight sparsity
    """

    def __init__(self, pruning_criterion=None, pruning_type=None, pruning_engine=enums.PruningEngine.PyTorch):
        self._validate_pruning_type(pruning_type)
        self._validate_pruning_engine(pruning_engine)

        self.pruning_type = pruning_type
        self.pruning_engine = pruning_engine
        self.pruning_criterion = pruning_criterion

    def prune(self, *args, **kwargs):
        raise NotImplementedError

    def sync_prune(self, *args, **kwargs):
        raise NotImplementedError

    def set_pruning_criterion(self, pruning_criterion):
        self._validate_pruning_criterion(pruning_criterion)
        self.pruning_criterion = pruning_criterion

    def _validate_pruning_type(self, pruning_type):
        assert pruning_type in (enums.PruningType.Structured, enums.PruningType.Unstructured)

    def _validate_pruning_engine(self, pruning_engine):
        assert pruning_engine in (enums.PruningEngine.PyTorch, enums.PruningEngine.NumPy)

    def _validate_pruning_criterion(self, pruning_criterion):
        assert isinstance(pruning_criterion, prune.pruning_criterion.BasePruningCriterion)
        assert self.pruning_engine == pruning_criterion.pruning_engine

    def _validate_pruning_amount(self, amount):
        assert isinstance(amount, float) and 0 <= amount < 1

    def _validate_importance_scores(self, weight, importance_scores):
        if self.pruning_type == enums.PruningType.Unstructured:
            assert weight.shape == importance_scores.shape
        elif self.pruning_type == enums.PruningType.Structured:
            assert weight.shape[0] == importance_scores.shape[0]

    def _validate_identical_nfilters(self, weights_pt):
        assert isinstance(weights_pt, list)
        # assert all(isinstance(module, torch.nn.Module) for module in modules)
        assert all(isinstance(weight_pt, torch.Tensor) for weight_pt in weights_pt)
        nfilters = self._get_nfilters(weights_pt[0])
        assert all(nfilters == self._get_nfilters(weight_pt) for weight_pt in weights_pt)

    def _validate_identical_amounts(self, amounts):
        assert isinstance(amounts, list)
        assert (all(isinstance(amount, float) for amount in amounts) or
                all(isinstance(amount, int) for amount in amounts))
        assert all(amounts[0] == amount for amount in amounts)

    def _t(self, tensor):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return tensor.detach()
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return tensor.detach().cpu().numpy()

    def _u(self, tensor, device):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return tensor
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return torch.from_numpy(tensor).to(device)

    def _transpose_io(self, module, tensor):
        if module:
            if module != 'cat_op' and not isinstance(module, torch.nn.Module):  # 如果既不是cat_op的虚拟权重也不是Module真实权重
                raise AssertionError
            if isinstance(module, torch.nn.ConvTranspose1d) or \
                    isinstance(module, torch.nn.ConvTranspose2d) or \
                    isinstance(module, torch.nn.ConvTranspose3d):
                tensor = torch.transpose(tensor, 0, 1)
        return tensor

    def _apply_mask(self, weight, mask):
        assert weight.shape == mask.shape
        weight_masked = weight * mask
        return weight_masked

    def _get_nfilters(self, tensor):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return tensor.shape[0]
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return tensor.shape[0]

    def _get_nweights(self, tensor):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return tensor.numel()
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return tensor.size

    def _get_nweights_toprune(self, nweights, amount):
        if isinstance(amount, float):
            self._validate_pruning_amount(amount)
            nweights_toprune = int(math.floor(nweights * amount))
        elif isinstance(amount, int):
            self._validate_pruning_amount(amount / nweights)
            nweights_toprune = amount
        return nweights_toprune

    def _bottomk(self, importance_scores, amount):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return torch.topk(importance_scores, amount, largest=False)[1]
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return np.argpartition(importance_scores, amount)[:amount]

    def _ones_like(self, tensor, dtype, device=None):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return torch.ones_like(tensor, dtype=dtype, device=device if device else tensor.device)
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return np.ones_like(tensor, dtype=dtype)

    def _zeros_like(self, tensor, dtype, device=None):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return torch.zeros_like(tensor, dtype=dtype, device=device if device else tensor.device)
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return np.zeros_like(tensor, dtype=dtype)

    def _ones(self, shape, dtype, device=None):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return torch.ones(shape, dtype=dtype, device=device)
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return np.ones(shape, dtype=dtype)

    def _zeros(self, shape, dtype, device=None):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return torch.zeros(shape, dtype=dtype, device=device)
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return np.zeros(shape, dtype=dtype)

    def _transpose_matrix(self, matrix):
        if self.pruning_engine == enums.PruningEngine.PyTorch:
            return torch.t(matrix)
        elif self.pruning_engine == enums.PruningEngine.NumPy:
            return np.transpose(matrix)


class StructuredPruningMethod(BasePruningMethod):
    __doc__ = r"""Structured pruning method for weight sparsity.
    """

    def __init__(self, pruning_criterion=None, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_criterion=pruning_criterion, pruning_type=enums.PruningType.Structured,
                         pruning_engine=pruning_engine)


class UnstructuredPruningMethod(BasePruningMethod):
    __doc__ = r"""Unstructured pruning method for weight sparsity.
    """

    def __init__(self, pruning_criterion=None, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_criterion=pruning_criterion, pruning_type=enums.PruningType.Unstructured,
                         pruning_engine=pruning_engine)
