import math
import sys
sys.path.append('../../..')
from knnslim.utils import enums
from knnslim.prune.pruning_method import base


class BCRPruningMethod(base.UnstructuredPruningMethod):
    __doc__ = r"""Block-based column-row (BCR) pruning method for unstructured weight sparsity.
    Args:
        block_h (int): Height of block
        block_w (int): Width of block
        prune_row (bool): To prune rows in block
        prune_col (bool): To prune cols in block
        pruning_criterion (enum.Enum): the criterion to calculate the importance of columns and rows
        pruning_engine (enum.Enum): PyTorch or NumPy, the engine for weight sparsity
    """
    def __init__(self, block_h=8, block_w=8, prune_row=True, prune_col=True, pruning_criterion=None, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_criterion=pruning_criterion, pruning_engine=pruning_engine)
        self.block_h = block_h
        self.block_w = block_w
        self.prune_row = prune_row
        self.prune_col = prune_col

    def prune(self, module, weight_pt, amount, criterion, importance_scores=None):
        weight_pt = self._transpose_io(module, weight_pt)

        device = weight_pt.device
        weight = self._t(weight_pt)

        if criterion and criterion != self.pruning_criterion:
            self.set_pruning_criterion(criterion)

        # get the number of pruning weights
        nweights = self._get_nweights(weight)
        nweights_toprune = self._get_nweights_toprune(nweights, amount)

        weight_matrix = weight.reshape(weight.shape[0], -1)
        weight_h, weight_w = weight_matrix.shape

        nblocks_row = int(math.ceil(weight_h / self.block_h))
        nblocks_col = int(math.ceil(weight_w / self.block_w))

        # calculate the importance scores of columns and rows in blocks and record their coordinates
        block_col_row_list = []

        for row in range(nblocks_row):
            for col in range(nblocks_col):
                weight_block = weight_matrix[row*self.block_h:(row+1)*self.block_h, 
                                             col*self.block_w:(col+1)*self.block_w]
                for dim in range(2):
                    if dim == 1:
                        weight_block = self._transpose_matrix(weight_block)
                    importance_scores_block = self.pruning_criterion(weight=weight_block)
                    for idx, score in enumerate(importance_scores_block):
                        block_col_row_list.append((row, col, dim, idx, score))

        # sort columns and rows according to their importance scores from smallest to largest
        block_col_row_list.sort(key=lambda record: record[-1])

        # record the numbers of pruned weights, pruned columns, and pruned rows
        nweights_pruned = 0
        ncols_pruned = self._zeros((nblocks_row, nblocks_col), dtype=int, device=device)
        nrows_pruned = self._zeros((nblocks_row, nblocks_col), dtype=int, device=device)

        # obtain the pruning mask and prune the columns and rows
        mask = self._ones((weight_h, weight_w), dtype=bool, device=device)

        for (row, col, dim, idx, score) in block_col_row_list:
            if nweights_pruned >= nweights_toprune:
                break
            if dim == 0 and self.prune_row:     # prune the [idx]-th row in the block at (row, col)
                nrows_pruned[row][col] += 1
                mask[row * self.block_h + idx, col * self.block_w:(col + 1) * self.block_w] = False
                nweights_pruned += self.block_w - ncols_pruned[row][col]
            elif dim == 1 and self.prune_col:   # prune the [idx]-th col in the block at (row, col)
                ncols_pruned[row][col] += 1
                mask[row * self.block_h:(row + 1) * self.block_h, col * self.block_w + idx] = False
                nweights_pruned += self.block_h - nrows_pruned[row][col]

        mask = mask.reshape(weight_pt.shape)
        weight_masked = self._apply_mask(weight, mask)

        weight_masked_pt = self._u(weight_masked, device)
        mask_pt = self._u(mask, device)

        weight_masked_pt = self._transpose_io(module, weight_masked_pt)
        mask_pt = self._transpose_io(module, mask_pt)
        return weight_masked_pt, mask_pt
