import sys
sys.path.append('../../..')

import torch

from knnslim.utils import enums
from knnslim.prune.pruning_method import filter


class ChannelPruningMethod(filter.FilterPruningMethod):
    __doc__ = r"""Channel pruning method for structured weight sparsity inherited from the filter pruning method.
    Args:
        pruning_criterion (enum.Enum): the criterion to calculate the importance of channels
        pruning_engine (enum.Enum): PyTorch or NumPy, the engine for weight sparsity
    """
    def __init__(self, pruning_criterion=None, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_criterion=pruning_criterion, pruning_engine=pruning_engine)

    def prune(self, module, weight_pt, amount, criterion, importance_scores=None):
        transposed_weight_pt = torch.transpose(weight_pt, 0, 1)
        transposed_weight_masked_pt, transposed_mask_pt = super().prune(module, transposed_weight_pt, amount, criterion, importance_scores)
        weight_masked_pt = torch.transpose(transposed_weight_masked_pt, 0, 1)
        mask_pt = torch.transpose(transposed_mask_pt, 0, 1)
        return weight_masked_pt, mask_pt

    def sync_prune(self, modules, weights_pt, amounts, criterion, importance_scores=None):
        transposed_weights_pt = [torch.transpose(weight_pt, 0, 1) for weight_pt in weights_pt]
        transposed_weights_masked_pt, transposed_masks_pt = super().sync_prune(modules, transposed_weights_pt, amounts, criterion, importance_scores)
        weights_masked_pt = [torch.transpose(weight_pt, 0, 1) for weight_pt in transposed_weights_masked_pt]
        masks_pt = [torch.transpose(transposed_mask_pt, 0, 1) for transposed_mask_pt in transposed_masks_pt]
        return weights_masked_pt, masks_pt
    
    def union_prune(self, modules, weights_pt, amounts, criterion, importance_scores=None, union_mask=False):
        transposed_weights_pt = [torch.transpose(weight_pt, 0, 1) for weight_pt in weights_pt]
        if union_mask:
            transposed_weights_masked_pt, transposed_masks_pt, union_mask_pt = super().union_prune(modules, transposed_weights_pt, amounts, criterion, importance_scores, union_mask=True)
        else:
            transposed_weights_masked_pt, transposed_masks_pt = super().union_prune(modules, transposed_weights_pt, amounts, criterion, importance_scores, union_mask=False)
        weights_masked_pt = [torch.transpose(weight_pt, 0, 1) for weight_pt in transposed_weights_masked_pt]
        masks_pt = [torch.transpose(transposed_mask_pt, 0, 1) for transposed_mask_pt in transposed_masks_pt]
        if union_mask:
            return weights_masked_pt, masks_pt, union_mask_pt
        return weights_masked_pt, masks_pt