import sys
sys.path.append('../../..')
from functools import reduce
from operator import add

import torch

from knnslim.utils import enums
from knnslim.prune.pruning_method import base


class FilterPruningMethod(base.StructuredPruningMethod):
    __doc__ = r"""Filter pruning method for structured weight sparsity.
    Args:
        pruning_criterion (enum.Enum): the criterion to calculate the importance of filters
        pruning_engine (enum.Enum): PyTorch or NumPy, the engine for weight sparsity
    """
    def __init__(self, pruning_criterion=None, pruning_engine=enums.PruningEngine.PyTorch):
        super().__init__(pruning_criterion=pruning_criterion, pruning_engine=pruning_engine)

    def prune(self, module, weight_pt, amount, criterion=None, importance_scores=None):
        """Prune the input weight_pt (PyTorch Tensor)
        Args:
            module (torch.nn.Module): module to prune, corresponding to the weight.
            weight_pt (torch.Tensor): weight to prune.
            amount (float or int): quantity of weights/filters/channels to prune.
            criterion (class, optional): criterion to calculate importance of weights/filters/channels.
            importance_scores (torch.Tensor, optional): importance of weights/filters/channels.
        Return:
            weight_masked_pt (torch.Tensor): weight masked/pruned
            mask_pt (torch.Tensor): mask tensor
        """
        weight_pt = self._transpose_io(module, weight_pt) # 翻转ConvTranspose层的参数
        device = weight_pt.device
        weight = self._t(weight_pt)

        # get the number of pruning filters
        nfilters = self._get_nfilters(weight)
        nfilters_toprune = self._get_nweights_toprune(nfilters, amount)

        # calculate the importance scores of all filters
        if not importance_scores:
            if criterion and criterion != self.pruning_criterion:
                self.set_pruning_criterion(criterion)
            importance_scores = self.pruning_criterion(weight=weight)
        self._validate_importance_scores(weight, importance_scores)

        # obtain the index of pruning filters
        pruning_filters = self._bottomk(importance_scores, nfilters_toprune)

        # obtain the pruning mask and prune the weight
        mask = self._ones_like(weight, dtype=bool)
        mask[pruning_filters] = False
        weight_masked = self._apply_mask(weight, mask)

        weight_masked_pt = self._u(weight_masked, device)
        mask_pt = self._u(mask, device)

        weight_masked_pt = self._transpose_io(module, weight_masked_pt)
        mask_pt = self._transpose_io(module, mask_pt)
        return weight_masked_pt, mask_pt

    def sync_prune(self, modules, weights_pt, amounts, criterion=None, importance_scores=None):
        """Prune synchronously the inputs weights_pt (list of PyTorch Tensors)
        Args:
            modules (list of torch.nn.Module): module list to prune, corresponding to the weight.
            weights_pt (list of torch.Tensor): weight list to prune.
            amounts (list of float or int): quantity list of weights/filters/channels to prune.
            criterion (class, optional): criterion to calculate importance of weights/filters/channels.
            importance_scores (torch.Tensor, optional): importance of weights/filters/channels.
        Return:
            weights_masked_pt (list of torch.Tensor): weight list masked/pruned
            masks_pt (list of torch.Tensor): mask tensor list
        """
        weights_pt = [self._transpose_io(module, weight_pt) for module, weight_pt in zip(modules, weights_pt)]

        self._validate_identical_nfilters(weights_pt)
        self._validate_identical_amounts(amounts)

        device = weights_pt[0].device
        weights = [self._t(weight_pt) for weight_pt in weights_pt]

        # get the number of pruning filters
        nfilters = self._get_nfilters(weights[0])
        nfilters_toprune = self._get_nweights_toprune(nfilters, amounts[0])

        # calculate the importance scores of all filters across all weights
        if not importance_scores:
            if criterion and criterion != self.pruning_criterion:
                self.set_pruning_criterion(criterion)
            importance_scores_list = [self.pruning_criterion(weight=weight) for weight in weights]
            importance_scores = reduce(add, importance_scores_list)
        self._validate_importance_scores(weights[0], importance_scores)

        # obtain the index of pruning filters
        pruning_filters = self._bottomk(importance_scores, nfilters_toprune)

        # obtain the pruning mask and prune the weight
        weights_masked = []
        masks = []
        for weight in weights:
            mask = self._ones_like(weight, dtype=bool)
            mask[pruning_filters] = False
            masks.append(mask)
            weights_masked.append(self._apply_mask(weight, mask))

        weights_masked_pt = [self._u(weight_masked, device) for weight_masked in weights_masked]
        masks_pt = [self._u(mask, device) for mask in masks]

        weights_masked_pt = [self._transpose_io(module, weight_masked) for module, weight_masked in zip(modules, weights_masked_pt)]
        masks_pt = [self._transpose_io(module, mask_pt) for module, mask_pt in zip(modules, masks_pt)]
        return weights_masked_pt, masks_pt
    
    def union_prune(self, modules, weights_pt, amounts, criterion=None, importance_scores=None, union_mask=False):
        """Prune independently the inputs weights_pt (list of PyTorch Tensors) and union the independent masks
        Args:
            modules (list of torch.nn.Module): module list to prune, corresponding to the weight.
            weights_pt (list of torch.Tensor): weight list to prune.
            amounts (list of float or int): quantity list of weights/filters/channels to prune.
            criterion (class, optional): criterion to calculate importance of weights/filters/channels.
            importance_scores (torch.Tensor, optional): importance of weights/filters/channels.
            union_mask (book): return union mask or not
        Return:
            weights_masked_pt (list of torch.Tensor): weight list masked/pruned
            masks_pt (list of torch.Tensor): mask tensor list
            union_mask (torch.Tensor, optional): union mask tensor
        """
        weights_masked_pt = []
        masks_pt = []
        for module, weight_pt, amount in zip(modules, weights_pt, amounts):
            weight_masked_pt, mask_pt = self.prune(module, weight_pt, amount, criterion, importance_scores)
            weights_masked_pt.append(weight_masked_pt)
            masks_pt.append(mask_pt)
        if union_mask:
            union_mask_pt = reduce(lambda a,b: torch.where(a>0,a,b), masks_pt)
            return weights_masked_pt, masks_pt, union_mask_pt
        return weights_masked_pt, masks_pt